package tcp

import (
	"bufio"
	"fmt"
	"log"
	"net"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

type TcpDaemon struct {
	daemon.Daemon

	Timeout        time.Duration
	Namespace      string
	ServiceType    string
	Keepalive      bool
	CollectService func(*TcpDaemon, *common.Service, *common.MetricData, string)

	connections map[string]net.Conn
	services    []*common.Service
}

func (d *TcpDaemon) ID() string {
	if len(d.ServiceType) > 0 {
		return d.ServiceType
	}

	return "tcp"
}

func (d *TcpDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)
	return nil
}

func (d *TcpDaemon) OnServicesChange(services common.ServicesMap) error {
	d.reload(services)
	return nil
}

func (d *TcpDaemon) reload(services common.ServicesMap) {
	if len(d.ServiceType) == 0 || d.CollectService == nil {
		return
	}

	// close open connexions
	for _, conn := range d.connections {
		_ = conn.Close()
	}

	// instanciate new connections list if keepalive is desired
	if d.Keepalive {
		d.connections = make(map[string]net.Conn)
	}
	d.services = services[d.ServiceType]
}

func (d *TcpDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for _, svc := range d.services {
		wg.Add(1)
		go func(service *common.Service) {
			defer wg.Done()
			target := fmt.Sprintf("%s:%d", service.IP, service.Port)

			d.CollectService(d, service, m, target)
		}(svc)
	}
	wg.Wait()

	return nil
}

func (d *TcpDaemon) Command(svc *common.Service, cmd string, ends []string) ([]string, error) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	// retrieve connection from cache
	conn, ok := d.connections[target]
	if ok {
		data, err := d.write(conn, cmd, ends)
		if err == nil {
			return data, err
		}
		log.Printf("WARN connection from cache reported error, closing and reopening (%s)", err)
		conn.Close()
		delete(d.connections, target)
	}

	conn, err := net.DialTimeout("tcp", target, d.Timeout)
	if err != nil {
		return nil, err
	}
	if !d.Keepalive {
		defer conn.Close()
	}

	data, err := d.write(conn, cmd, ends)
	if d.Keepalive && err == nil {
		d.connections[target] = conn
	}
	return data, err
}

func (d *TcpDaemon) write(conn net.Conn, cmd string, ends []string) ([]string, error) {
	if err := conn.SetDeadline(time.Now().Add(d.Timeout)); err != nil {
		err = fmt.Errorf("Error setting timeout writing command '%s' to '%s' (%w)", cmd, conn.RemoteAddr(), err)
		return nil, err
	}
	if _, err := conn.Write([]byte(cmd)); err != nil {
		err = fmt.Errorf("Error writing command '%s' to '%s' (%w)", cmd, conn.RemoteAddr(), err)
		return nil, err
	}

	data := []string{}

	scanner := bufio.NewScanner(conn)
	for scanner.Scan() {
		line := scanner.Text()
		for _, end := range ends {
			if line == end {
				return data, nil
			}
		}
		data = append(data, line)
	}
	return data, nil
}
