package daemon

import (
	"context"
	"fmt"
	"log"
	"time"

	"github.com/open-io/oio-exporter/common"
)

// TODO: add context to every method
type DaemonI interface {
	ID() string
	Interval() time.Duration
	Init(common.ServicesMap, []*common.Service) error
	Collect(*common.MetricData) error
	Close(context.Context) error
}

type ServicesDaemonI interface {
	OnServicesChange(common.ServicesMap) error
}

type TargetsDaemonI interface {
	OnTargetsChange([]*common.Service) error
}

type Daemon struct{}

func (d *Daemon) Interval() time.Duration {
	return 5 * time.Second
}

func (d *Daemon) Init(common.ServicesMap, []*common.Service) error {
	return nil
}

func (d *Daemon) Close(ctx context.Context) error {
	return nil
}

func Run(d DaemonI, m *common.MetricData, state *common.State, stopC <-chan struct{}) error {
	daemonID := d.ID()
	interval := d.Interval()

	services := state.Services
	targets := state.Targets

	if err := d.Init(services, targets); err != nil {
		return fmt.Errorf("Init %s: %w", daemonID, err)
	}

	servicesC := make(chan common.ServicesMap)
	targetsC := make(chan []*common.Service)
	timer := time.AfterFunc(time.Hour, func() {
		log.Printf("WARN %s collection is taking longer than its interval of %s", daemonID, interval)
	})
	ticker := time.NewTicker(interval)
	defer ticker.Stop()

	// In case of a long init
	select {
	case <-stopC:
		goto exit
	default:
	}

	if d, ok := d.(ServicesDaemonI); ok {
		go onServicesChange(d, state, state.ServicesLastUpdate, servicesC, stopC)
	}

	if d, ok := d.(TargetsDaemonI); ok {
		go onTargetsChange(d, state, state.TargetsLastUpdate, targetsC, stopC)
	}

	timer.Reset(interval)
	if err := d.Collect(m); err != nil {
		log.Printf("ERROR Collecting %s: %s", daemonID, err)
	}
	timer.Stop()

	for {
		// This is required
		// If ticker.C is already ready by the time the loop starts again
		// The second select's behavior wouldn't be determined
		select {
		case <-stopC:
			goto exit
		default:
		}

		select {
		case <-stopC:
			goto exit
		case services := <-servicesC:
			log.Printf("Reloading services for %s", daemonID)

			if err := d.(ServicesDaemonI).OnServicesChange(services); err != nil {
				log.Printf("WARN Error while reloading '%s': %s", daemonID, err)
			}

			continue
		case targets := <-targetsC:
			log.Printf("Reloading targets for %s", daemonID)

			if err := d.(TargetsDaemonI).OnTargetsChange(targets); err != nil {
				log.Printf("WARN Error while reloading '%s': %s", daemonID, err)
			}

			continue
		case <-ticker.C:
		}

		timer.Reset(interval)
		if err := d.Collect(m); err != nil {
			log.Printf("ERROR Collecting %s: %s", daemonID, err)
		}
		timer.Stop()
	}

exit:
	if err := d.Close(context.TODO()); err != nil {
		return fmt.Errorf("Closing %s: %w", daemonID, err)
	}

	return nil
}

func onServicesChange(d ServicesDaemonI, state *common.State, lastUpdate time.Time, servicesC chan<- common.ServicesMap, stopC <-chan struct{}) {
	daemonID := d.(DaemonI).ID()

	for {
		state.OnServicesChange.L.Lock()
		for state.ServicesLastUpdate.Sub(lastUpdate) == 0 {
			log.Printf("Waiting for services change %s", daemonID)
			state.OnServicesChange.Wait()

			select {
			case <-stopC:
				state.OnServicesChange.L.Unlock()
				return
			default:
			}
		}

		lastUpdate = state.ServicesLastUpdate
		services := state.Services

		state.OnServicesChange.L.Unlock()

		servicesC <- services
	}
}

func onTargetsChange(d TargetsDaemonI, state *common.State, lastUpdate time.Time, targetsC chan<- []*common.Service, stopC <-chan struct{}) {
	daemonID := d.(DaemonI).ID()

	for {
		state.OnTargetsChange.L.Lock()
		for state.TargetsLastUpdate.Sub(lastUpdate) == 0 {
			log.Printf("Waiting for targets change %s", daemonID)
			state.OnTargetsChange.Wait()

			select {
			case <-stopC:
				state.OnTargetsChange.L.Unlock()
				return
			default:
			}
		}

		lastUpdate = state.TargetsLastUpdate
		targets := state.Targets

		state.OnTargetsChange.L.Unlock()

		targetsC <- targets
	}
}
